
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"

#include "meta-enum-types.h"
#include "meta-theme.h"

GType
meta_theme_error_get_type (void)
{
  static gsize g_enum_type_id;
  static const GEnumValue values[] =
    {
      { META_THEME_ERROR_TOO_OLD, "META_THEME_ERROR_TOO_OLD", "too-old" },
      { META_THEME_ERROR_FRAME_GEOMETRY, "META_THEME_ERROR_FRAME_GEOMETRY", "frame-geometry" },
      { META_THEME_ERROR_BAD_CHARACTER, "META_THEME_ERROR_BAD_CHARACTER", "bad-character" },
      { META_THEME_ERROR_BAD_PARENS, "META_THEME_ERROR_BAD_PARENS", "bad-parens" },
      { META_THEME_ERROR_UNKNOWN_VARIABLE, "META_THEME_ERROR_UNKNOWN_VARIABLE", "unknown-variable" },
      { META_THEME_ERROR_DIVIDE_BY_ZERO, "META_THEME_ERROR_DIVIDE_BY_ZERO", "divide-by-zero" },
      { META_THEME_ERROR_MOD_ON_FLOAT, "META_THEME_ERROR_MOD_ON_FLOAT", "mod-on-float" },
      { META_THEME_ERROR_FAILED, "META_THEME_ERROR_FAILED", "failed" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaThemeError");
      id = g_enum_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

GType
meta_theme_type_get_type (void)
{
  static gsize g_enum_type_id;
  static const GEnumValue values[] =
    {
      { META_THEME_TYPE_GTK, "META_THEME_TYPE_GTK", "gtk" },
      { META_THEME_TYPE_METACITY, "META_THEME_TYPE_METACITY", "metacity" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaThemeType");
      id = g_enum_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

/* Generated data ends here */

