
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef META_ENUM_TYPES_H
#define META_ENUM_TYPES_H

#include <glib-object.h>

G_BEGIN_DECLS
GType meta_window_type_get_type (void) G_GNUC_CONST;
#define META_TYPE_WINDOW_TYPE (meta_window_type_get_type())
GType meta_maximize_flags_get_type (void) G_GNUC_CONST;
#define META_TYPE_MAXIMIZE_FLAGS (meta_maximize_flags_get_type())
GType meta_client_type_get_type (void) G_GNUC_CONST;
#define META_TYPE_CLIENT_TYPE (meta_client_type_get_type())
GType meta_queue_type_get_type (void) G_GNUC_CONST;
#define META_TYPE_QUEUE_TYPE (meta_queue_type_get_type())
GType meta_compositor_type_get_type (void) G_GNUC_CONST;
#define META_TYPE_COMPOSITOR_TYPE (meta_compositor_type_get_type())
G_END_DECLS

#endif

/* Generated data ends here */

