
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "config.h"
#include "meta-enum-types.h"
#include "window-private.h"

GType
meta_window_type_get_type (void)
{
  static gsize g_enum_type_id;
  static const GEnumValue values[] =
    {
      { META_WINDOW_NORMAL, "META_WINDOW_NORMAL", "normal" },
      { META_WINDOW_DESKTOP, "META_WINDOW_DESKTOP", "desktop" },
      { META_WINDOW_DOCK, "META_WINDOW_DOCK", "dock" },
      { META_WINDOW_DIALOG, "META_WINDOW_DIALOG", "dialog" },
      { META_WINDOW_MODAL_DIALOG, "META_WINDOW_MODAL_DIALOG", "modal-dialog" },
      { META_WINDOW_TOOLBAR, "META_WINDOW_TOOLBAR", "toolbar" },
      { META_WINDOW_MENU, "META_WINDOW_MENU", "menu" },
      { META_WINDOW_UTILITY, "META_WINDOW_UTILITY", "utility" },
      { META_WINDOW_SPLASHSCREEN, "META_WINDOW_SPLASHSCREEN", "splashscreen" },
      { META_WINDOW_DROPDOWN_MENU, "META_WINDOW_DROPDOWN_MENU", "dropdown-menu" },
      { META_WINDOW_POPUP_MENU, "META_WINDOW_POPUP_MENU", "popup-menu" },
      { META_WINDOW_TOOLTIP, "META_WINDOW_TOOLTIP", "tooltip" },
      { META_WINDOW_NOTIFICATION, "META_WINDOW_NOTIFICATION", "notification" },
      { META_WINDOW_COMBO, "META_WINDOW_COMBO", "combo" },
      { META_WINDOW_DND, "META_WINDOW_DND", "dnd" },
      { META_WINDOW_OVERRIDE_OTHER, "META_WINDOW_OVERRIDE_OTHER", "override-other" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaWindowType");
      id = g_enum_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

GType
meta_maximize_flags_get_type (void)
{
  static gsize g_enum_type_id;
  static const GFlagsValue values[] =
    {
      { META_MAXIMIZE_HORIZONTAL, "META_MAXIMIZE_HORIZONTAL", "horizontal" },
      { META_MAXIMIZE_VERTICAL, "META_MAXIMIZE_VERTICAL", "vertical" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaMaximizeFlags");
      id = g_flags_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

GType
meta_client_type_get_type (void)
{
  static gsize g_enum_type_id;
  static const GEnumValue values[] =
    {
      { META_CLIENT_TYPE_UNKNOWN, "META_CLIENT_TYPE_UNKNOWN", "unknown" },
      { META_CLIENT_TYPE_APPLICATION, "META_CLIENT_TYPE_APPLICATION", "application" },
      { META_CLIENT_TYPE_PAGER, "META_CLIENT_TYPE_PAGER", "pager" },
      { META_CLIENT_TYPE_MAX_RECOGNIZED, "META_CLIENT_TYPE_MAX_RECOGNIZED", "max-recognized" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaClientType");
      id = g_enum_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

GType
meta_queue_type_get_type (void)
{
  static gsize g_enum_type_id;
  static const GFlagsValue values[] =
    {
      { META_QUEUE_CALC_SHOWING, "META_QUEUE_CALC_SHOWING", "calc-showing" },
      { META_QUEUE_MOVE_RESIZE, "META_QUEUE_MOVE_RESIZE", "move-resize" },
      { META_QUEUE_UPDATE_ICON, "META_QUEUE_UPDATE_ICON", "update-icon" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaQueueType");
      id = g_flags_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}
#include "meta-compositor.h"

GType
meta_compositor_type_get_type (void)
{
  static gsize g_enum_type_id;
  static const GEnumValue values[] =
    {
      { META_COMPOSITOR_TYPE_NONE, "META_COMPOSITOR_TYPE_NONE", "none" },
      { META_COMPOSITOR_TYPE_XRENDER, "META_COMPOSITOR_TYPE_XRENDER", "xrender" },
      { META_COMPOSITOR_TYPE_XPRESENT, "META_COMPOSITOR_TYPE_XPRESENT", "xpresent" },
      { 0, NULL, NULL }
    };

  if (g_once_init_enter (&g_enum_type_id))
    {
      const gchar *string;
      GType id;

      string = g_intern_static_string ("MetaCompositorType");
      id = g_enum_register_static (string, values);

      g_once_init_leave (&g_enum_type_id, id);
    }

  return g_enum_type_id;
}

/* Generated data ends here */

